import json
import os
import random
import sys
import time
from typing import Dict, Optional, Callable, Any

import numpy as np
import tensorflow as tf
from tensorflow.python.training.tracking import data_structures as tf_data_structures
from dpu_utils.utils import RichPath

from ..data import DataFold, GraphDataset
from ..layers import get_known_message_passing_classes
from ..models import GraphTaskModel
from .model_utils import save_model, load_weights_verbosely, get_model_and_dataset
from .task_utils import get_known_tasks


def make_run_id(model_name: str, task_name: str, run_name: Optional[str] = None) -> str:
    """Choose a run ID, based on the --run-name parameter and the current time."""
    if run_name is not None:
        return run_name
    else:
        return "%s_%s__%s" % (model_name, task_name, time.strftime("%Y-%m-%d_%H-%M-%S"))


def log_line(log_file: str, msg: str):
    with open(log_file, "a") as log_fh:
        log_fh.write(msg + "\n")
    print(msg)


def train_loop(
    model: GraphTaskModel,
    train_data: tf.data.Dataset,
    valid_data: tf.data.Dataset,
    max_epochs: int,
    patience: int,
    log_fun: Callable[[str], None],
    save_model_fun: Callable[[GraphTaskModel], None],
    quiet: bool = False,
    aml_run=None,
) -> float:
    _, _, initial_valid_results = model.run_one_epoch(valid_data, training=False, quiet=quiet)
    best_valid_metric, best_val_str = model.compute_epoch_metrics(initial_valid_results)
    log_fun(f"Initial valid metric: {best_val_str}.")
    save_model_fun(model)
    best_valid_epoch = 0
    train_time_start = time.time()

    for epoch in range(1, max_epochs + 1):
        log_fun(f"== Epoch {epoch}")
        train_loss, train_speed, train_results = model.run_one_epoch(
            train_data, training=True, quiet=quiet
        )
        train_metric, train_metric_string = model.compute_epoch_metrics(train_results)
        log_fun(
            f" Train:  {train_loss:.4f} loss | {train_metric_string} | {train_speed:.2f} graphs/s",
        )
        valid_loss, valid_speed, valid_results = model.run_one_epoch(
            valid_data, training=False, quiet=quiet
        )
        valid_metric, valid_metric_string = model.compute_epoch_metrics(valid_results)
        log_fun(
            f" Valid:  {valid_loss:.4f} loss | {valid_metric_string} | {valid_speed:.2f} graphs/s",
        )

        if aml_run is not None:
            aml_run.log("task_train_metric", float(train_metric))
            aml_run.log("train_speed", float(train_speed))
            aml_run.log("task_valid_metric", float(valid_metric))
            aml_run.log("valid_speed", float(valid_speed))

        # Save if good enough.
        if valid_metric < best_valid_metric:
            log_fun(
                f"  (Best epoch so far, target metric decreased to {valid_metric:.5f} from {best_valid_metric:.5f}.)",
            )
            save_model_fun(model)
            best_valid_metric = valid_metric
            best_valid_epoch = epoch
        elif epoch - best_valid_epoch >= patience:
            total_time = time.time() - train_time_start
            log_fun(
                f"Stopping training after {patience} epochs without "
                f"improvement on validation metric.",
            )
            log_fun(f"Training took {total_time}s. Best validation metric: {best_valid_metric}",)
            break

    return best_valid_metric


def train(
    model: GraphTaskModel,
    dataset: GraphDataset,
    log_fun: Callable[[str], None],
    run_id: str,
    max_epochs: int,
    patience: int,
    save_dir: str,
    quiet: bool = False,
    aml_run=None,
):
    train_data = dataset.get_tensorflow_dataset(DataFold.TRAIN).prefetch(3)
    valid_data = dataset.get_tensorflow_dataset(DataFold.VALIDATION).prefetch(3)

    save_file = os.path.join(save_dir, f"{run_id}_best.pkl")

    def save_model_fun(model: GraphTaskModel):
        save_model(save_file, model, dataset)

    train_loop(
        model,
        train_data,
        valid_data,
        max_epochs=max_epochs,
        patience=patience,
        log_fun=log_fun,
        save_model_fun=save_model_fun,
        quiet=quiet,
        aml_run=aml_run,
    )

    return save_file


def unwrap_tf_tracked_data(data: Any) -> Any:
    if isinstance(data, (tf_data_structures.ListWrapper, list)):
        return [unwrap_tf_tracked_data(e) for e in data]
    elif isinstance(data, (tf_data_structures._DictWrapper, dict)):
        return {k: unwrap_tf_tracked_data(v) for k, v in data.items()}
    else:
        return data


def run_train_from_args(args, hyperdrive_hyperparameter_overrides: Dict[str, str] = {}) -> None:
    # Get the housekeeping going and start logging:
    os.makedirs(args.save_dir, exist_ok=True)
    run_id = make_run_id(args.model, args.task, args.run_name)
    log_file = os.path.join(args.save_dir, f"{run_id}.log")
    
    def log(msg):
        log_line(log_file, msg)

    log(f"Setting random seed {args.random_seed}.")
    random.seed(args.random_seed)
    np.random.seed(args.random_seed)
    tf.random.set_seed(args.random_seed)

    data_path = RichPath.create(args.data_path, args.azure_info)
    dataset, model = get_model_and_dataset(
        msg_passing_implementation=args.model,
        task_name=args.task,
        data_path=data_path,
        trained_model_file=args.load_saved_model,
        cli_data_hyperparameter_overrides=args.data_param_override,
        cli_model_hyperparameter_overrides=args.model_param_override,
        hyperdrive_hyperparameter_overrides=hyperdrive_hyperparameter_overrides,
        folds_to_load={DataFold.TRAIN, DataFold.VALIDATION},
        load_weights_only=args.load_weights_only,
    )

    log(f"Dataset parameters: {json.dumps(unwrap_tf_tracked_data(dataset._params))}")
    log(f"Model parameters: {json.dumps(unwrap_tf_tracked_data(model._params))}")

    if args.azureml_logging:
        from azureml.core.run import Run

        aml_run = Run.get_context()
    else:
        aml_run = None

    trained_model_path = train(
        model,
        dataset,
        log_fun=log,
        run_id=run_id,
        max_epochs=args.max_epochs,
        patience=args.patience,
        save_dir=args.save_dir,
        quiet=args.quiet,
        aml_run=aml_run,
    )

    if args.run_test:
        data_path = RichPath.create(args.data_path, args.azure_info)
        log("== Running on test dataset")
        log(f"Loading data from {data_path}.")
        dataset.load_data(data_path, {DataFold.TEST})
        log(f"Restoring best model state from {trained_model_path}.")
        load_weights_verbosely(trained_model_path, model)

        # Test 1: Simply compute same metrics used during training/validation:
        test_data = dataset.get_tensorflow_dataset(DataFold.TEST)
        _, _, test_results = model.run_one_epoch(test_data, training=False, quiet=args.quiet)
        test_metric, test_metric_string = model.compute_epoch_metrics(test_results)
        log(test_metric_string)
        if aml_run is not None:
            aml_run.log("task_test_metric", float(test_metric))

        # Test 2: Try to compute fancier metrics, if implemented:
        try:
            eval_metrics = model.evaluate_model(test_data)
            for metric_name, metric_value in eval_metrics.items():
                log(f"{metric_name:<30}: {metric_value:8.4f}")
                if aml_run is not None:
                    aml_run.log(f"task_test_{metric_name}", metric_value)
        except NotImplementedError:
            pass  # ignore if there are no fancier metrics


def get_train_cli_arg_parser(default_model_type: Optional[str]=None):
    """
    Get an argparse argument parser object with common options for training
    GNN-based models.

    Args:
        default_model_type: If provided, the model type is downgraded from a
            positional parameter on the command line to an option with the
            given default value.
    """
    import argparse

    parser = argparse.ArgumentParser(description="Train a GNN model.")
    # We use a somewhat horrible trick to support both
    #  train.py --model MODEL --task TASK --data_path DATA_PATH
    # as well as
    #  train.py model task data_path
    # The former is useful because of limitations in AzureML; the latter is nicer to type.
    if "--task" in sys.argv:
        model_param_name, task_param_name, data_path_param_name = "--model", "--task", "--data_path"
    else:
        model_param_name, task_param_name, data_path_param_name = "model", "task", "data_path"

    if default_model_type:
        model_param_name = "--model"
    parser.add_argument(
        model_param_name,
        type=str,
        choices=sorted(get_known_message_passing_classes()),
        default=default_model_type,
        help="GNN model type to train.",
    )
    parser.add_argument(
        task_param_name,
        type=str,
        choices=sorted(get_known_tasks()),
        help="Task to train model for.",
    )
    parser.add_argument(data_path_param_name, type=str, help="Directory containing the task data.")
    parser.add_argument(
        "--save-dir",
        dest="save_dir",
        type=str,
        default="trained_model",
        help="Path in which to store the trained model and log.",
    )
    parser.add_argument(
        "--model-params-override",
        dest="model_param_override",
        type=str,
        help="JSON dictionary overriding model hyperparameter values.",
    )
    parser.add_argument(
        "--data-params-override",
        dest="data_param_override",
        type=str,
        help="JSON dictionary overriding data hyperparameter values.",
    )
    parser.add_argument(
        "--max-epochs",
        dest="max_epochs",
        type=int,
        default=10000,
        help="Maximal number of epochs to train for.",
    )
    parser.add_argument(
        "--patience",
        dest="patience",
        type=int,
        default=25,
        help="Maximal number of epochs to continue training without improvement.",
    )
    parser.add_argument(
        "--seed", dest="random_seed", type=int, default=0, help="Random seed to use.",
    )
    parser.add_argument(
        "--run-name", dest="run_name", type=str, help="A human-readable name for this run.",
    )
    parser.add_argument(
        "--azure-info",
        dest="azure_info",
        type=str,
        default="azure_auth.json",
        help="Azure authentication information file (JSON).",
    )
    parser.add_argument(
        "--load-saved-model",
        dest="load_saved_model",
        help="Optional location to load initial model weights from. Should be model stored in earlier run.",
    )
    parser.add_argument(
        "--load-weights-only",
        dest="load_weights_only",
        action="store_true",
        help="Optional to only load the weights of the model rather than class and dataset for further training (used in fine-tuning on pretrained network). Should be model stored in earlier run.",
    )
    parser.add_argument(
        "--quiet", dest="quiet", action="store_true", help="Generate less output during training.",
    )
    parser.add_argument(
        "--run-test",
        dest="run_test",
        action="store_true",
        default=False,
        help="Run on testset after training.",
    )
    parser.add_argument(
        "--azureml_logging",
        dest="azureml_logging",
        action="store_true",
        help="Log task results using AML run context.",
    )
    parser.add_argument("--debug", dest="debug", action="store_true", help="Enable debug routines")

    parser.add_argument(
        "--hyperdrive-arg-parse",
        dest="hyperdrive_arg_parse",
        action="store_true",
        help='Enable hyperdrive argument parsing, in which unknown options "--key val" are interpreted as hyperparameter "key" with value "val".',
    )

    return parser



"""Gated graph neural network layer."""
from typing import Dict, List, Any

import tensorflow as tf

from .message_passing import MessagePassing, MessagePassingInput, register_message_passing_implementation
from .gnn_edge_mlp import GNN_Edge_MLP
from tf2_gnn.utils.constants import SMALL_NUMBER


@register_message_passing_implementation
class GGNN(GNN_Edge_MLP):
    """Compute new graph states by neural message passing and gated units on the nodes.
    For this, we assume existing node states h^t_v and a list of per-edge-type adjacency
    matrices A_\ell.

    We compute new states as follows:
        h^{t+1}_v := Cell(h^t_v, \sum_\ell
                                 \sum_{(u, v) \in A_\ell}
                                     W_\ell * h^t_u)
    The learnable parameters of this are the recurrent Cell and the W_\ell \in R^{D,D}.

    We use the following abbreviations in shape descriptions:
    * V: number of nodes
    * L: number of different edge types
    * E: number of edges of a given edge type
    * D: input node representation dimension
    * H: output node representation dimension (set as hidden_dim)

    NOTE: in this layer, the dimension of the node embedding must be equal to the hidden dimension!

    >>> node_embeddings = tf.random.normal(shape=(5, 12))
    >>> adjacency_lists = (
    ...    tf.constant([[0, 1], [2, 4], [2, 4]], dtype=tf.int32),
    ...    tf.constant([[2, 3], [2, 4]], dtype=tf.int32),
    ...    tf.constant([[3, 1]], dtype=tf.int32),
    ... )
    ...
    >>> params = GGNN.get_default_hyperparameters()
    >>> params["hidden_dim"] = 12
    >>> layer = GGNN(params)
    >>> output = layer(MessagePassingInput(node_embeddings, adjacency_lists))
    >>> print(output)
    tf.Tensor(..., shape=(5, 12), dtype=float32)
    """

    @classmethod
    def get_default_hyperparameters(cls):
        these_hypers = {
            "use_target_state_as_input": False,
            "normalize_by_num_incoming": True,
            "num_edge_MLP_hidden_layers": 0,
        }
        mp_hypers = super().get_default_hyperparameters()
        mp_hypers.update(these_hypers)
        return mp_hypers

    def __init__(self, params: Dict[str, Any], **kwargs):
        super().__init__(params, **kwargs)
        self._recurrent_unit: tf.keras.layers.GRUCell = None

    def build(self, input_shapes: MessagePassingInput):
        node_embedding_shapes = input_shapes.node_embeddings
        self._recurrent_unit = tf.keras.layers.GRUCell(units=self._hidden_dim)
        self._recurrent_unit.build(tf.TensorShape((None, node_embedding_shapes[-1])))
        super().build(input_shapes)

    def _compute_new_node_embeddings(
        self,
        cur_node_embeddings: tf.Tensor,
        messages_per_type: List[tf.Tensor],
        edge_type_to_message_targets: List[tf.Tensor],
        num_nodes: tf.Tensor,
        training: bool,
    ):
        # Let M be the number of messages (sum of all E):
        message_targets = tf.concat(edge_type_to_message_targets, axis=0)  # Shape [M]
        messages = tf.concat(messages_per_type, axis=0)  # Shape [M, H]

        aggregated_messages = self._aggregation_fn(
            data=messages, segment_ids=message_targets, num_segments=num_nodes
        )

        new_node_embeddings, _ = self._recurrent_unit(
            inputs=aggregated_messages,
            states=[cur_node_embeddings],
            training=training)

        return new_node_embeddings

if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import os
import sys

import tensorflow as tf
from dpu_utils.utils import run_and_debug

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", ".."))


from tf2_gnn.cli_utils import get_train_cli_arg_parser, run_train_from_args


def run():
    parser = get_train_cli_arg_parser()
    args, potential_hyperdrive_args = parser.parse_known_args()

    hyperdrive_hyperparameter_overrides = None
    if args.hyperdrive_arg_parse and len(potential_hyperdrive_args) % 2 == 0:
        # Allow parsing params specified as "--key value" as well as "key value"
        hyperdrive_hyperparameter_overrides = {
            param.replace("--", ""): value
            for param, value in zip(potential_hyperdrive_args[::2], potential_hyperdrive_args[1::2])
        }
    elif len(potential_hyperdrive_args) > 0:
        # Reparse to throw standard error message:
        args = parser.parse_args()

    # Make TF less noisy:
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"
    tf.get_logger().setLevel("ERROR")

    run_and_debug(
        lambda: run_train_from_args(args, hyperdrive_hyperparameter_overrides), args.debug
    )


if __name__ == "__main__":
    run()



"""Functions to convert from string parameters to their values."""
import json
from typing import List, Dict, Any
from distutils.util import strtobool


def to_bool(val) -> bool:
    """ Accepts a boolean or str value, and returns the boolean equivalent, converting if necessary """

    if type(val) == bool:
        return val
    else:
        return bool(strtobool(val))


def str_to_list_of_ints(val) -> List[int]:
    """Accepts a str or list, returns list of ints. Specifically useful when 
    num_hidden_units of a set of layers is specified as a list of ints"""
    if type(val) == list:
        return val
    else:

        return [int(v) for v in json.loads(val)]


def override_model_params_with_hyperdrive_params(
    model_params: Dict[str, Any], hyperdrive_params: Dict[str, str]
):
    """
    Overrides the model parameters, with those from hyperdrive_params. hyperdrive_params contains hyperparameter values as strings. 
	The correct type is inferred from the model params (only the value is used from hyperdrive_params)
    """
    for k in hyperdrive_params.keys():
        if k not in model_params:
            raise ValueError(f"key {k} not found in model_params: {model_params}")

        if type(model_params[k]) == bool:
            model_params[k] = to_bool(hyperdrive_params[k])
        elif type(model_params[k]) == int:
            model_params[k] = int(hyperdrive_params[k])
        elif type(model_params[k]) == float:
            model_params[k] = float(hyperdrive_params[k])
        elif type(model_params[k]) == list and type(model_params[k][0]) == int:
            model_params[k] = str_to_list_of_ints(hyperdrive_params[k])
        elif type(model_params[k]) == str:
            model_params[k] = hyperdrive_params[k]
        else:
            raise ValueError(f"Unknown hyperparameter type {type(model_params[k])} for hyperparameter {k}.")
    return



"""Task for regression on QM9 dataset."""
from typing import Any, Dict, List, Tuple, Optional, Union

import tensorflow as tf
from dpu_utils.tf2utils import MLP

from tf2_gnn.data import GraphDataset, QM9Dataset
from tf2_gnn.models import GraphTaskModel


# These magic constants were obtained during dataset generation, as result of normalising
# the values of target properties:
CHEMICAL_ACC_NORMALISING_FACTORS = [
    0.066513725,
    0.012235489,
    0.071939046,
    0.033730778,
    0.033486113,
    0.004278493,
    0.001330901,
    0.004165489,
    0.004128926,
    0.00409976,
    0.004527465,
    0.012292586,
    0.037467458,
]


class QM9RegressionTask(GraphTaskModel):
    @classmethod
    def get_default_hyperparameters(
        cls, mp_style: Optional[str] = None
    ) -> Dict[str, Any]:
        super_params = super().get_default_hyperparameters(mp_style)
        these_hypers: Dict[str, Any] = {
            "use_intermediate_gnn_results": False,
            "out_layer_dropout_keep_prob": 1.0,
        }
        super_params.update(these_hypers)
        return super_params

    def __init__(self, params: Dict[str, Any], dataset: GraphDataset, name: str = None):
        super().__init__(params, dataset=dataset, name=name)
        assert isinstance(dataset, QM9Dataset)

        self._task_id = int(dataset._params["task_id"])

        self._regression_gate = MLP(
            out_size=1,
            hidden_layers=[],
            use_biases=True,
            dropout_rate=self._params["out_layer_dropout_keep_prob"],
            name="gate",
        )
        self._regression_transform = MLP(
            out_size=1,
            hidden_layers=[],
            use_biases=True,
            dropout_rate=self._params["out_layer_dropout_keep_prob"],
            name="transform"
        )

    def build(self, input_shapes):
        with tf.name_scope(self.__class__.__name__):
            with tf.name_scope("node_gate"):
                self._regression_gate.build(
                    tf.TensorShape(
                        (
                            None,
                            input_shapes["node_features"][-1]
                            + self._params["gnn_hidden_dim"],
                        )
                    )
                )
            with tf.name_scope("node_transform"):
                self._regression_transform.build(
                    tf.TensorShape((None, self._params["gnn_hidden_dim"]))
                )

        super().build(input_shapes)

    def compute_task_output(
        self,
        batch_features: Dict[str, tf.Tensor],
        final_node_representations: Union[tf.Tensor, Tuple[tf.Tensor, List[tf.Tensor]]],
        training: bool,
    ) -> Any:
        if self._params["use_intermediate_gnn_results"]:
            final_node_representations, _ = final_node_representations

        # The per-node regression uses only final node representations:
        per_node_output = self._regression_transform(
            final_node_representations, training=training
        )  # Shape [V, 1]

        # The gating uses both initial and final node representations:
        per_node_weight = self._regression_gate(
            tf.concat(
                [batch_features["node_features"], final_node_representations], axis=-1
            ),
            training=training,
        )  # Shape [V, 1]

        per_node_weighted_output = tf.squeeze(
            tf.nn.sigmoid(per_node_weight) * per_node_output, axis=-1
        )  # Shape [V]
        per_graph_output = tf.math.unsorted_segment_sum(
            data=per_node_weighted_output,
            segment_ids=batch_features["node_to_graph_map"],
            num_segments=batch_features["num_graphs_in_batch"],
        )  # Shape [G]

        return per_graph_output

    def compute_task_metrics(
        self,
        batch_features: Dict[str, tf.Tensor],
        task_output: Any,
        batch_labels: Dict[str, tf.Tensor],
    ) -> Dict[str, tf.Tensor]:
        mse = tf.losses.mean_squared_error(batch_labels["target_value"], task_output)
        mae = tf.losses.mean_absolute_error(batch_labels["target_value"], task_output)
        num_graphs = tf.cast(batch_features["num_graphs_in_batch"], tf.float32)
        return {
            "loss": mse,
            "batch_squared_error": mse * num_graphs,
            "batch_absolute_error": mae * num_graphs,
            "num_graphs": num_graphs,
        }

    def compute_epoch_metrics(self, task_results: List[Any]) -> Tuple[float, str]:
        total_num_graphs = sum(
            batch_task_result["num_graphs"] for batch_task_result in task_results
        )
        total_absolute_error = sum(
            batch_task_result["batch_absolute_error"]
            for batch_task_result in task_results
        )
        total_squared_error = sum(
            batch_task_result["batch_squared_error"]
            for batch_task_result in task_results
        )
        epoch_mse = (total_squared_error / total_num_graphs).numpy()
        epoch_mae = (total_absolute_error / total_num_graphs).numpy()
        return (
            epoch_mae,
            (
                f"Task {self._task_id} |"
                f" MSE = {epoch_mse:.3f} |"
                f" MAE = {epoch_mae:.3f} |"
                f" Error Ratio: {epoch_mae / CHEMICAL_ACC_NORMALISING_FACTORS[self._task_id]:.3f}"
            ),
        )



"""Relation graph attention network layer."""
from typing import Dict, List, Tuple, Any

import tensorflow as tf
from dpu_utils.tf2utils import unsorted_segment_log_softmax

from .message_passing import MessagePassing, MessagePassingInput, register_message_passing_implementation


@register_message_passing_implementation
class RGAT(MessagePassing):
    """Compute new graph states by neural message passing using attention. This generalises
    the original GAT model (Velickovic et al., https://arxiv.org/pdf/1710.10903.pdf)
    to multiple edge types by using different weights for different edge types.
    For this, we assume existing node states h^t_v and a list of per-edge-type adjacency
    matrices A_\ell.

    In the setting for a single attention head, we compute new states as follows:
        h^t_{v, \ell} := W_\ell h^t_v
        e_{u, \ell, v} := LeakyReLU(\alpha_\ell^T * concat(h^t_{u, \ell}, h^t_{v, \ell}))
        a_v := softmax_{\ell, u with (u, v) \in A_\ell}(e_{u, \ell, v})
        h^{t+1}_v := \sigma(\sum_{ell, (u, v) \in A_\ell}
                                a_v_{u, \ell} * h^_{u, \ell})
    The learnable parameters of this are the W_\ell \in R^{D, D} and \alpha_\ell \in R^{2*D}.

    In practice, we use K attention heads, computing separate, partial new states h^{t+1}_{v,k}
    and compute h^{t+1}_v as the concatentation of the partial states.
    For this, we reduce the shape of W_\ell to R^{D, D/K} and \alpha_\ell to R^{2*D/K}.

    We use the following abbreviations in shape descriptions:
    * V: number of nodes
    * K: number of attention heads
    * L: number of different edge types
    * E: number of edges of a given edge type
    * D: input node representation dimension
    * H: output node representation dimension (set as hidden_dim)

    >>> node_embeddings = tf.random.normal(shape=(5, 3))
    >>> adjacency_lists = (
    ...    tf.constant([[0, 1], [2, 4], [2, 4]], dtype=tf.int32),
    ...    tf.constant([[2, 3], [2, 4]], dtype=tf.int32),
    ...    tf.constant([[3, 1]], dtype=tf.int32),
    ... )
    ...
    >>> params = RGAT.get_default_hyperparameters()
    >>> params["hidden_dim"] = 12
    >>> layer = RGAT(params)
    >>> output = layer(MessagePassingInput(node_embeddings, adjacency_lists))
    >>> print(output)
    tf.Tensor(..., shape=(5, 12), dtype=float32)
    """

    @classmethod
    def get_default_hyperparameters(cls):
        these_hypers = {
            "num_heads": 3,
        }
        mp_hypers = super().get_default_hyperparameters()
        mp_hypers.update(these_hypers)
        return mp_hypers

    def __init__(self, params: Dict[str, Any], **kwargs):
        super().__init__(params, **kwargs)
        self._num_heads: int = params["num_heads"]
        self._edge_type_to_message_computation_layer: List[tf.keras.layers.Layer] = []
        self._edge_type_to_attention_parameters: List[tf.Variable] = []

    def build(self, input_shapes: MessagePassingInput):
        node_embedding_shapes = input_shapes.node_embeddings
        adjacency_list_shapes = input_shapes.adjacency_lists
        num_edge_types = len(adjacency_list_shapes)
        per_head_dim = self._hidden_dim // self._num_heads

        for i in range(num_edge_types):
            with tf.name_scope(f"edge_type_{i}"):
                mp_layer = tf.keras.layers.Dense(
                    self._hidden_dim, use_bias=False, name="Edge_weight_{}".format(i)
                )
                mp_layer.build(tf.TensorShape((None, node_embedding_shapes[-1])))
                self._edge_type_to_message_computation_layer.append(mp_layer)

                attention_weights = self.add_weight(
                    name="Edge_attention_parameters_{}".format(i),
                    shape=(self._num_heads, 2 * per_head_dim),
                    trainable=True,
                )
                self._edge_type_to_attention_parameters.append(attention_weights)

        super().build(input_shapes)

    def _message_function(
        self,
        edge_source_states: tf.Tensor,
        edge_target_states: tf.Tensor,
        num_incoming_to_node_per_message: tf.Tensor,
        edge_type_idx: int,
        training: bool,
    ) -> tf.Tensor:
        per_head_dim = self._hidden_dim // self._num_heads

        # Actually do the message calculation:
        per_head_transformed_source_states = tf.reshape(
            self._edge_type_to_message_computation_layer[edge_type_idx](edge_source_states),
            shape=(-1, self._num_heads, per_head_dim),
        )  # Shape [E, K, H/K]
        per_head_transformed_target_states = tf.reshape(
            self._edge_type_to_message_computation_layer[edge_type_idx](edge_target_states),
            shape=(-1, self._num_heads, per_head_dim),
        )  # Shape [E, K, H/K]

        per_head_transformed_states = tf.concat(
            [per_head_transformed_source_states, per_head_transformed_target_states], axis=-1
        )  # Shape [E, K, 2*H/K]

        per_head_attention_scores = tf.nn.leaky_relu(
            tf.einsum(
                "vki,ki->vk",
                per_head_transformed_states,
                self._edge_type_to_attention_parameters[edge_type_idx],
            )
        )  # Shape [E, K]

        return (per_head_transformed_source_states, per_head_attention_scores)

    def _compute_new_node_embeddings(
        self,
        cur_node_embeddings: tf.Tensor,
        messages_per_type: List[Tuple[tf.Tensor, tf.Tensor]],
        edge_type_to_message_targets: List[tf.Tensor],
        num_nodes: tf.Tensor,
        training: bool,
    ):
        per_head_messages_per_type, per_head_attention_scores_per_type = zip(*messages_per_type)

        per_head_messages = tf.concat(per_head_messages_per_type, axis=0)  # Shape [M, K, H/K]
        per_head_attention_scores = tf.concat(
            per_head_attention_scores_per_type, axis=0
        )  # Shape [M, K]
        message_targets = tf.concat(edge_type_to_message_targets, axis=0)  # Shape [M]

        head_to_aggregated_messages = []  # list of tensors of shape [V, H/K]
        for head_idx in range(self._num_heads):
            # Compute the softmax over all the attention scores for all messages going to this state:
            attention_scores = tf.concat(
                per_head_attention_scores[:, head_idx], axis=0
            )  # Shape [M]
            attention_values = tf.exp(
                unsorted_segment_log_softmax(
                    logits=attention_scores, segment_ids=message_targets, num_segments=num_nodes
                )
            )  # Shape [M]
            messages = per_head_messages[:, head_idx, :]  # Shape [M, H/K]
            # Compute weighted sum per target node for this head:
            head_to_aggregated_messages.append(
                tf.math.unsorted_segment_sum(
                    data=tf.expand_dims(attention_values, -1) * messages,
                    segment_ids=message_targets,
                    num_segments=num_nodes,
                )
            )

        aggregated_messages = tf.concat(head_to_aggregated_messages, axis=-1)  # Shape [V, H]
        return self._activation_fn(aggregated_messages)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



#!/usr/bin/env python3
import os
from typing import Callable

import tensorflow as tf
from dpu_utils.utils import run_and_debug, RichPath

from tf2_gnn import DataFold, GraphDataset, GraphTaskModel
from tf2_gnn.cli_utils.training_utils import get_model_and_dataset


def test(
    model: GraphTaskModel,
    dataset: GraphDataset,
    log_fun: Callable[[str], None],
    quiet: bool = False,
):
    log_fun("== Running on test dataset")
    test_data = dataset.get_tensorflow_dataset(DataFold.TEST)
    _, _, test_results = model.run_one_epoch(test_data, training=False, quiet=quiet)
    test_metric, test_metric_string = model.compute_epoch_metrics(test_results)
    log_fun(test_metric_string)


def run_from_args(args) -> None:
    data_path = RichPath.create(args.DATA_PATH, args.azure_info)
    dataset, model = get_model_and_dataset(
        msg_passing_implementation=None,
        task_name=None,
        data_path=data_path,
        trained_model_file=args.TRAINED_MODEL,
        cli_data_hyperparameter_overrides=args.data_param_override,
        cli_model_hyperparameter_overrides=args.model_param_override,
        folds_to_load={DataFold.TEST},
    )
    test(model, dataset, lambda msg: print(msg), quiet=args.quiet)


def run():
    import argparse

    parser = argparse.ArgumentParser(description="Test a GNN model.")
    parser.add_argument(
        "TRAINED_MODEL",
        type=str,
        help="File to load model from (determines model architecture & task).",
    )
    parser.add_argument("DATA_PATH", type=str, help="Directory containing the task data.")
    parser.add_argument(
        "--model-params-override",
        dest="model_param_override",
        type=str,
        help="JSON dictionary overriding model hyperparameter values.",
    )
    parser.add_argument(
        "--data-params-override",
        dest="data_param_override",
        type=str,
        help="JSON dictionary overriding data hyperparameter values.",
    )
    parser.add_argument(
        "--azure-info",
        dest="azure_info",
        type=str,
        default="azure_auth.json",
        help="Azure authentication information file (JSON).",
    )
    parser.add_argument(
        "--quiet", dest="quiet", action="store_true", help="Generate less output during testing.",
    )
    parser.add_argument("--debug", dest="debug", action="store_true", help="Enable debug routines")
    args = parser.parse_args()

    # Shut up tensorflow:
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'
    tf.get_logger().setLevel("ERROR")
    import warnings
    warnings.simplefilter("ignore")

    run_and_debug(lambda: run_from_args(args), args.debug)


if __name__ == "__main__":
    run()



from .graph_dataset import DataFold, GraphSample, GraphBatchTFDataDescription, GraphDataset
from .jsonl_graph_dataset import JsonLGraphDataset
from .jsonl_graph_property_dataset import JsonLGraphPropertyDataset
from .ppi_dataset import PPIDataset
from .qm9_dataset import QM9Dataset



from typing import Union

import tensorflow as tf


class PolynomialWarmupAndDecaySchedule(
    tf.keras.optimizers.schedules.LearningRateSchedule
):
    """A LearningRateSchedule that uses a polynomial warmup and decay schedule."""

    def __init__(
        self,
        learning_rate: Union[tf.Tensor, float],
        warmup_steps: Union[tf.Tensor, int],
        decay_steps: Union[tf.Tensor, int],
        initial_learning_rate: Union[tf.Tensor, float],
        final_learning_rate: Union[tf.Tensor, float],
        power: Union[tf.Tensor, float] = 1.0,
        name=None,
    ):
        """Applies polynomial warmup and decay to the learning rate.
        In the warmup phase of the first `warmup_steps` training steps, the learning
        rate is increased from `initial_learning_rate` to `learning_rate`.
        In the following `decay_steps` steps, the learning rate is decreased to
        `final_learning_rate`.

        The schedule is a 1-arg callable that produces a decayed learning rate
        when passed the current optimizer step. This can be useful for changing the
        learning rate value across different invocations of optimizer functions.
        It is computed as:
        ```python
        def scheduled_learning_rate(step):
            if step <= warmup_steps:
                return ((learning_rate - initial_learning_rate) *
                        (step / warmup_steps)**power
                    ) + initial_learning_rate
            else:
                step = min(step - warmup_steps, decay_steps)
                return ((learning_rate - final_learning_rate) *
                        (1 - step / decay_steps)**power
                    ) + final_learning_rate
        ```

        You can pass this schedule directly into a `tf.keras.optimizers.Optimizer`
        as the learning rate.

        Args:
            learning_rate: A scalar `float32` or `float64` `Tensor` or a
                Python number.  The peak learning rate.
            warmup_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
                Must be positive.  See the warmup computation above.
            decay_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
                Must be positive.  See the decay computation above.
            initial_learning_rate: A scalar `float32` or `float64` `Tensor` or a
                Python number, setting the initial learning rate from which we
                warm up to `learning_rate`.
            final_learning_rate: A scalar `float32` or `float64` `Tensor` or a
                Python number, setting the final learning rate to which we decay
                from `learning_rate`.
            power: A scalar `float32` or `float64` `Tensor` or a
                Python number.  The power of the polynomial. Defaults to linear, 1.0.
            name: String.  Optional name of the operation.

        Returns:
            A 1-arg callable learning rate schedule that takes the current optimizer
            step and outputs the decayed learning rate, a scalar `Tensor` of the same
            type as `initial_learning_rate`.
        """
        super(PolynomialWarmupAndDecaySchedule, self).__init__()

        self.learning_rate = learning_rate
        self.initial_learning_rate = initial_learning_rate
        self.final_learning_rate = final_learning_rate
        self.warmup_steps = warmup_steps
        self.decay_steps = decay_steps
        self.power = power
        self.name = name

    def get_config(self):
        return {
            "learning_rate": self.learning_rate,
            "initial_learning_rate": self.initial_learning_rate,
            "final_learning_rate": self.final_learning_rate,
            "warmup_steps": self.warmup_steps,
            "decay_steps": self.decay_steps,
            "power": self.power,
            "name": self.name,
        }

    def __call__(self, step):
        with tf.name_scope(self.name or "PolynomialWarmupAndDecay"):
            is_in_warmup_phase = tf.less_equal(step, self.warmup_steps)

            def _warmup_learning_rate():
                return (
                    (self.learning_rate - self.initial_learning_rate)
                    * tf.pow(step / self.warmup_steps, self.power)
                ) + self.initial_learning_rate

            def _decay_learning_rate():
                effective_step = tf.math.minimum(step - self.warmup_steps, self.decay_steps)
                return (
                    (self.learning_rate - self.final_learning_rate)
                    * tf.pow(1 - effective_step / self.decay_steps, self.power)
                ) + self.final_learning_rate

            return tf.cond(
                is_in_warmup_phase,
                true_fn=_warmup_learning_rate,
                false_fn=_decay_learning_rate,
            )



"""General task for GNN regression."""
from typing import Any, Dict, List, Tuple, Optional, Union

import tensorflow as tf
from dpu_utils.tf2utils import MLP

from tf2_gnn.data import GraphDataset
from tf2_gnn.models import GraphTaskModel
from tf2_gnn.layers import (
    WeightedSumGraphRepresentation,
    NodesToGraphRepresentationInput,
)


class GraphRegressionTask(GraphTaskModel):
    @classmethod
    def get_default_hyperparameters(
        cls, mp_style: Optional[str] = None
    ) -> Dict[str, Any]:
        super_params = super().get_default_hyperparameters(mp_style)
        these_hypers: Dict[str, Any] = {
            "use_intermediate_gnn_results": True,
            "graph_aggregation_output_size": 32,
            "graph_aggregation_num_heads": 4,
            "graph_aggregation_layers": [32, 32],
            "graph_aggregation_dropout_rate": 0.1,
            "regression_mlp_layers": [64, 32],
            "regression_mlp_dropout": 0.1,
        }
        super_params.update(these_hypers)
        return super_params

    def __init__(self, params: Dict[str, Any], dataset: GraphDataset, name: str = None):
        super().__init__(params, dataset=dataset, name=name)
        self._node_to_graph_aggregation = None

        # Construct sublayers:
        self._weighted_avg_of_nodes_to_graph_repr = WeightedSumGraphRepresentation(
            graph_representation_size=self._params["graph_aggregation_output_size"],
            num_heads=self._params["graph_aggregation_num_heads"],
            weighting_fun="softmax",
            scoring_mlp_layers=self._params["graph_aggregation_layers"],
            scoring_mlp_dropout_rate=self._params["graph_aggregation_dropout_rate"],
            scoring_mlp_activation_fun="elu",
            transformation_mlp_layers=self._params["graph_aggregation_layers"],
            transformation_mlp_dropout_rate=self._params[
                "graph_aggregation_dropout_rate"
            ],
            transformation_mlp_activation_fun="elu",
        )
        self._weighted_sum_of_nodes_to_graph_repr = WeightedSumGraphRepresentation(
            graph_representation_size=self._params["graph_aggregation_output_size"],
            num_heads=self._params["graph_aggregation_num_heads"],
            weighting_fun="sigmoid",
            scoring_mlp_layers=self._params["graph_aggregation_layers"],
            scoring_mlp_dropout_rate=self._params["graph_aggregation_dropout_rate"],
            scoring_mlp_activation_fun="elu",
            transformation_mlp_layers=self._params["graph_aggregation_layers"],
            transformation_mlp_dropout_rate=self._params[
                "graph_aggregation_dropout_rate"
            ],
            transformation_mlp_activation_fun="elu",
        )

        self._regression_mlp = MLP(
            out_size=1,
            hidden_layers=self._params["regression_mlp_layers"],
            dropout_rate=self._params["regression_mlp_dropout"],
            use_biases=True,
            activation_fun=tf.nn.relu,
        )

    def build(self, input_shapes):
        if self._params["use_intermediate_gnn_results"]:
            # We get the initial GNN input + results for all layers:
            node_repr_size = (
                input_shapes["node_features"][-1]
                + self._params["gnn_hidden_dim"] * self._params["gnn_num_layers"]
            )
        else:
            node_repr_size = (
                input_shapes["node_features"][-1] + self._params["gnn_hidden_dim"]
            )

        node_to_graph_repr_input = NodesToGraphRepresentationInput(
            node_embeddings=tf.TensorShape((None, node_repr_size)),
            node_to_graph_map=tf.TensorShape((None)),
            num_graphs=tf.TensorShape(()),
        )

        with tf.name_scope(self.__class__.__name__):
            with tf.name_scope("graph_representation_computation"):
                with tf.name_scope("weighted_avg"):
                    self._weighted_avg_of_nodes_to_graph_repr.build(
                        node_to_graph_repr_input
                    )
                with tf.name_scope("weighted_sum"):
                    self._weighted_sum_of_nodes_to_graph_repr.build(
                        node_to_graph_repr_input
                    )

            self._regression_mlp.build(
                tf.TensorShape(
                    (None, 2 * self._params["graph_aggregation_output_size"])
                )
            )

        super().build(input_shapes)

    def compute_task_output(
        self,
        batch_features: Dict[str, tf.Tensor],
        final_node_representations: Union[tf.Tensor, Tuple[tf.Tensor, List[tf.Tensor]]],
        training: bool,
    ) -> Any:
        if self._params["use_intermediate_gnn_results"]:
            _, intermediate_node_representations = final_node_representations
            # We want to skip the first "intermediate" representation, which is the output of
            # the initial feature -> GNN input layer:
            node_representations = tf.concat(
                (batch_features["node_features"],)
                + intermediate_node_representations[1:],
                axis=-1,
            )
        else:
            node_representations = tf.concat(
                [batch_features["node_features"], final_node_representations], axis=-1
            )

        graph_representation_layer_input = NodesToGraphRepresentationInput(
            node_embeddings=node_representations,
            node_to_graph_map=batch_features["node_to_graph_map"],
            num_graphs=batch_features["num_graphs_in_batch"],
        )
        weighted_avg_graph_repr = self._weighted_avg_of_nodes_to_graph_repr(
            graph_representation_layer_input, training=training
        )
        weighted_sum_graph_repr = self._weighted_sum_of_nodes_to_graph_repr(
            graph_representation_layer_input, training=training
        )

        graph_representations = tf.concat(
            [weighted_avg_graph_repr, weighted_sum_graph_repr], axis=-1
        )  # shape: [G, GD]

        per_graph_results = self._regression_mlp(
            graph_representations, training=training
        )  # shape: [G, 1]

        return tf.squeeze(per_graph_results, axis=-1)

    def compute_task_metrics(
        self,
        batch_features: Dict[str, tf.Tensor],
        task_output: Any,
        batch_labels: Dict[str, tf.Tensor],
    ) -> Dict[str, tf.Tensor]:
        mse = tf.losses.mean_squared_error(batch_labels["target_value"], task_output)
        mae = tf.losses.mean_absolute_error(batch_labels["target_value"], task_output)
        num_graphs = tf.cast(batch_features["num_graphs_in_batch"], tf.float32)
        return {
            "loss": mse,
            "batch_squared_error": mse * num_graphs,
            "batch_absolute_error": mae * num_graphs,
            "num_graphs": num_graphs,
        }

    def compute_epoch_metrics(self, task_results: List[Any]) -> Tuple[float, str]:
        total_num_graphs = sum(
            batch_task_result["num_graphs"] for batch_task_result in task_results
        )
        total_absolute_error = sum(
            batch_task_result["batch_absolute_error"]
            for batch_task_result in task_results
        )
        total_squared_error = sum(
            batch_task_result["batch_squared_error"]
            for batch_task_result in task_results
        )
        epoch_mse = (total_squared_error / total_num_graphs).numpy()
        epoch_mae = (total_absolute_error / total_num_graphs).numpy()
        return epoch_mae, f" MSE = {epoch_mse:.3f} | MAE = {epoch_mae:.3f}"

    def evaluate_model(self, dataset: tf.data.Dataset) -> Dict[str, float]:
        import sklearn.metrics as metrics

        predictions = self.predict(dataset).numpy()
        labels = []
        for _, batch_labels in dataset:
            labels.append(batch_labels["target_value"])
        labels = tf.concat(labels, axis=0).numpy()

        metrics = dict(
            mae=metrics.mean_absolute_error(y_true=labels, y_pred=predictions),
            mse=metrics.mean_squared_error(y_true=labels, y_pred=predictions),
            max_err=metrics.max_error(y_true=labels, y_pred=predictions),
            expl_var=metrics.explained_variance_score(
                y_true=labels, y_pred=predictions
            ),
            r2_score=metrics.r2_score(y_true=labels, y_pred=predictions),
        )

        return metrics



"""Custom activation functions."""
import math as m

import tensorflow as tf


def gelu(input_tensor: tf.Tensor):
    """An approximation to the GELU activation function as used in the paper
    https://arxiv.org/pdf/1810.04805.pdf
    """
    cdf = 0.5 * (
        1.0 + tf.tanh((tf.sqrt(2 / m.pi) * (input_tensor + 0.044715 * tf.pow(input_tensor, 3))))
    )
    return input_tensor * cdf



from .layers import GNNInput, GNN, get_known_message_passing_classes
from .models import GraphTaskModel, NodeMulticlassTask, GraphRegressionTask, GraphBinaryClassificationTask
from .data import DataFold, GraphSample, GraphBatchTFDataDescription, GraphDataset, JsonLGraphDataset



from .dataset_utils import load_dataset_for_prediction
from .model_utils import load_model_for_prediction
from .task_utils import task_name_to_dataset_class, task_name_to_model_class, register_task, get_known_tasks, clear_known_tasks
from .training_utils import make_run_id, get_train_cli_arg_parser, run_train_from_args



"""Graph representation aggregation layer."""
from abc import abstractmethod
from typing import List, NamedTuple, Optional

import tensorflow as tf
from dpu_utils.tf2utils import MLP, get_activation_function_by_name, unsorted_segment_softmax


class NodesToGraphRepresentationInput(NamedTuple):
    """A named tuple to hold input to layers computing graph representations from nodes
    representations."""

    node_embeddings: tf.Tensor
    node_to_graph_map: tf.Tensor
    num_graphs: tf.Tensor


class NodesToGraphRepresentation(tf.keras.layers.Layer):
    """Abstract class to compute graph representations from node representations.

    Throughout we use the following abbreviations in shape descriptions:
        * V: number of nodes (across all graphs)
        * VD: node representation dimension
        * G: number of graphs
        * GD: graph representation dimension
    """

    def __init__(self, graph_representation_size: int, **kwargs):
        super().__init__(**kwargs)
        self._graph_representation_size = graph_representation_size

    @abstractmethod
    def call(self, inputs: NodesToGraphRepresentationInput, training: bool = False):
        """Call the layer.

        Args:
            inputs: A tuple containing two items:
                node_embeddings: float32 tensor of shape [V, VD], the representation of each
                    node in all graphs.
                node_to_graph_map: int32 tensor of shape [V] with values in range [0, G-1],
                    mapping each node to a graph ID.
                num_graphs: int32 scalar, specifying the number G of graphs.
            training: A bool that denotes whether we are in training mode.

        Returns:
            float32 tensor of shape [G, GD]
        """
        pass


class WeightedSumGraphRepresentation(NodesToGraphRepresentation):
    """Layer computing graph representations as weighted sum of node representations.
    The weights are either computed from the original node representations ("self-attentional")
    or by a softmax across the nodes of a graph.
    Supports splitting operation into parallely computed independent "heads" which can focus
    on different aspects.

    Throughout we use the following abbreviations in shape descriptions:
        * V: number of nodes (across all graphs)
        * VD: node representation dimension
        * G: number of graphs
        * GD: graph representation dimension
        * H: number of heads
    """

    def __init__(
        self,
        graph_representation_size: int,
        num_heads: int,
        weighting_fun: str = "softmax",  # One of {"softmax", "sigmoid"}
        scoring_mlp_layers: List[int] = [128],
        scoring_mlp_activation_fun: str = "ReLU",
        scoring_mlp_use_biases: bool = False,
        scoring_mlp_dropout_rate: float = 0.2,
        transformation_mlp_layers: List[int] = [128],
        transformation_mlp_activation_fun: str = "ReLU",
        transformation_mlp_use_biases: bool = False,
        transformation_mlp_dropout_rate: float = 0.2,
        transformation_mlp_result_lower_bound: Optional[float] = None,
        transformation_mlp_result_upper_bound: Optional[float] = None,
        **kwargs,
    ):
        """
        Args:
            graph_representation_size: Size of the computed graph representation.
            num_heads: Number of independent heads to use to compute weights.
            weighting_fun: "sigmoid" ([0, 1] weights for each node computed from its
                representation), "softmax" ([0, 1] weights for each node computed
                from all nodes in same graph), "average" (weight is fixed to 1/num_nodes),
                or "none" (weight is fixed to 1).
            scoring_mlp_layers: MLP layer structure for computing raw scores turned into
                weights.
            scoring_mlp_activation_fun: MLP activcation function for computing raw scores
                turned into weights.
            scoring_mlp_dropout_rate: MLP inter-layer dropout rate for computing raw scores
                turned into weights.
            transformation_mlp_layers: MLP layer structure for computing graph representations.
            transformation_mlp_activation_fun: MLP activcation function for computing graph
                representations.
            transformation_mlp_dropout_rate: MLP inter-layer dropout rate for computing graph
                representations.
            transformation_mlp_result_lower_bound: Lower bound that results of the transformation
                MLP will be clipped to before being scaled and summed up.
                This is particularly useful to limit the magnitude of results when using "sigmoid"
                or "none" as weighting function.
            transformation_mlp_result_upper_bound: Upper bound that results of the transformation
                MLP will be clipped to before being scaled and summed up.
        """
        super().__init__(graph_representation_size, **kwargs)
        assert (
            graph_representation_size % num_heads == 0
        ), f"Number of heads {num_heads} needs to divide final representation size {graph_representation_size}!"
        assert weighting_fun.lower() in {
            "none",
            "average",
            "softmax",
            "sigmoid",
        }, f"Weighting function {weighting_fun} unknown, {{'softmax', 'sigmoid', 'none', 'average'}} supported."

        self._num_heads = num_heads
        self._weighting_fun = weighting_fun.lower()
        self._transformation_mlp_activation_fun = get_activation_function_by_name(
            transformation_mlp_activation_fun
        )
        self._transformation_mlp_result_lower_bound = transformation_mlp_result_lower_bound
        self._transformation_mlp_result_upper_bound = transformation_mlp_result_upper_bound

        # Build sub-layers:
        if self._weighting_fun not in ("none", "average"):
            self._scoring_mlp = MLP(
                out_size=self._num_heads,
                hidden_layers=scoring_mlp_layers,
                use_biases=scoring_mlp_use_biases,
                activation_fun=get_activation_function_by_name(
                    scoring_mlp_activation_fun
                ),
                dropout_rate=scoring_mlp_dropout_rate,
                name="ScoringMLP",
            )

        self._transformation_mlp = MLP(
            out_size=self._graph_representation_size,
            hidden_layers=transformation_mlp_layers,
            use_biases=transformation_mlp_use_biases,
            activation_fun=self._transformation_mlp_activation_fun,
            dropout_rate=transformation_mlp_dropout_rate,
            name="TransformationMLP",
        )

    def build(self, input_shapes: NodesToGraphRepresentationInput):
        with tf.name_scope("WeightedSumGraphRepresentation"):
            if self._weighting_fun not in ("none", "average"):
                self._scoring_mlp.build(
                    tf.TensorShape((None, input_shapes.node_embeddings[-1]))
                )
            self._transformation_mlp.build(tf.TensorShape((None, input_shapes.node_embeddings[-1])))

            super().build(input_shapes)

    @tf.function(
        input_signature=(
            NodesToGraphRepresentationInput(
                node_embeddings=tf.TensorSpec(shape=tf.TensorShape((None, None)), dtype=tf.float32),
                node_to_graph_map=tf.TensorSpec(shape=tf.TensorShape((None,)), dtype=tf.int32),
                num_graphs=tf.TensorSpec(shape=(), dtype=tf.int32),
            ),
            tf.TensorSpec(shape=(), dtype=tf.bool),
        )
    )
    def call(self, inputs: NodesToGraphRepresentationInput, training: bool = False):
        # (1) compute weights for each node/head pair:
        if self._weighting_fun not in ("none", "average"):
            scores = self._scoring_mlp(inputs.node_embeddings, training=training)  # Shape [V, H]
            if self._weighting_fun == "sigmoid":
                weights = tf.nn.sigmoid(scores)  # Shape [V, H]
            elif self._weighting_fun == "softmax":
                weights_per_head = []
                for head_idx in range(self._num_heads):
                    head_scores = scores[:, head_idx]  # Shape [V]
                    head_weights = unsorted_segment_softmax(
                        logits=head_scores,
                        segment_ids=inputs.node_to_graph_map,
                        num_segments=inputs.num_graphs,
                    )  # Shape [V]
                    weights_per_head.append(tf.expand_dims(head_weights, -1))
                weights = tf.concat(weights_per_head, axis=1)  # Shape [V, H]
            else:
                raise ValueError()

        # (2) compute representations for each node/head pair:
        node_reprs = self._transformation_mlp_activation_fun(
            self._transformation_mlp(inputs.node_embeddings, training=training)
        )  # Shape [V, GD]
        if self._transformation_mlp_result_lower_bound is not None:
            node_reprs = tf.maximum(node_reprs, self._transformation_mlp_result_lower_bound)
        if self._transformation_mlp_result_upper_bound is not None:
            node_reprs = tf.minimum(node_reprs, self._transformation_mlp_result_upper_bound)
        node_reprs = tf.reshape(
            node_reprs,
            shape=(-1, self._num_heads, self._graph_representation_size // self._num_heads),
        )  # Shape [V, H, GD//H]

        # (3) if necessary, weight representations and aggregate by graph:
        if self._weighting_fun == "none":
            node_reprs = tf.reshape(
                node_reprs, shape=(-1, self._graph_representation_size)
            )  # Shape [V, GD]
            graph_reprs = tf.math.segment_sum(
                data=node_reprs, segment_ids=inputs.node_to_graph_map
            )  # Shape [G, GD]
        elif self._weighting_fun == "average":
            node_reprs = tf.reshape(
                node_reprs, shape=(-1, self._graph_representation_size)
            )  # Shape [V, GD]
            graph_reprs = tf.math.segment_mean(
                data=node_reprs, segment_ids=inputs.node_to_graph_map
            )  # Shape [G, GD]
        else:
            weights = tf.expand_dims(weights, -1)  # Shape [V, H, 1]
            weighted_node_reprs = weights * node_reprs  # Shape [V, H, GD//H]

            weighted_node_reprs = tf.reshape(
                weighted_node_reprs, shape=(-1, self._graph_representation_size)
            )  # Shape [V, GD]
            graph_reprs = tf.math.segment_sum(
                data=weighted_node_reprs, segment_ids=inputs.node_to_graph_map
            )  # Shape [G, GD]

        return graph_reprs



from .graph_task_model import GraphTaskModel
from .node_multiclass_task import NodeMulticlassTask
from .graph_regression_task import GraphRegressionTask
from .graph_binary_classification_task import GraphBinaryClassificationTask
from .qm9_regression import QM9RegressionTask


"""General task for graph binary classification."""
from typing import Any, Dict, List, Tuple, Optional

import numpy as np
import tensorflow as tf

from tf2_gnn.data import GraphDataset
from tf2_gnn.models.graph_regression_task import GraphRegressionTask


class GraphBinaryClassificationTask(GraphRegressionTask):
    @classmethod
    def get_default_hyperparameters(
        cls, mp_style: Optional[str] = None
    ) -> Dict[str, Any]:
        super_params = super().get_default_hyperparameters(mp_style)
        these_hypers: Dict[str, Any] = {}
        super_params.update(these_hypers)
        return super_params

    def __init__(self, params: Dict[str, Any], dataset: GraphDataset, name: str = None):
        super().__init__(params, dataset=dataset, name=name)

    def compute_task_output(
        self,
        batch_features: Dict[str, tf.Tensor],
        final_node_representations: tf.Tensor,
        training: bool,
    ) -> Any:
        per_graph_regression_results = super().compute_task_output(
            batch_features, final_node_representations, training
        )

        return tf.nn.sigmoid(per_graph_regression_results)

    def compute_task_metrics(
        self,
        batch_features: Dict[str, tf.Tensor],
        task_output: Any,
        batch_labels: Dict[str, tf.Tensor],
    ) -> Dict[str, tf.Tensor]:
        ce = tf.reduce_mean(
            tf.keras.losses.binary_crossentropy(
                y_true=batch_labels["target_value"],
                y_pred=task_output,
                from_logits=False,
            )
        )
        num_correct = tf.reduce_sum(
            tf.cast(
                tf.math.equal(batch_labels["target_value"], tf.math.round(task_output)),
                tf.int32,
            )
        )
        num_graphs = tf.cast(batch_features["num_graphs_in_batch"], tf.float32)
        return {
            "loss": ce,
            "batch_acc": tf.cast(num_correct, tf.float32) / num_graphs,
            "num_correct": num_correct,
            "num_graphs": num_graphs,
        }

    def compute_epoch_metrics(self, task_results: List[Any]) -> Tuple[float, str]:
        total_num_graphs = np.sum(
            batch_task_result["num_graphs"] for batch_task_result in task_results
        )
        total_num_correct = np.sum(
            batch_task_result["num_correct"] for batch_task_result in task_results
        )
        epoch_acc = tf.cast(total_num_correct, tf.float32) / total_num_graphs
        return -epoch_acc.numpy(), f"Accuracy = {epoch_acc.numpy():.3f}"

    def evaluate_model(self, dataset: tf.data.Dataset) -> Dict[str, float]:
        import sklearn.metrics as metrics

        predictions = self.predict(dataset).numpy()
        rounded_preds = np.round(predictions)
        labels = []
        for _, batch_labels in dataset:
            labels.append(batch_labels["target_value"])
        labels = tf.concat(labels, axis=0).numpy()

        metrics = dict(
            acc=metrics.accuracy_score(y_true=labels, y_pred=rounded_preds),
            balanced_acc=metrics.balanced_accuracy_score(
                y_true=labels, y_pred=rounded_preds
            ),
            precicision=metrics.precision_score(y_true=labels, y_pred=rounded_preds),
            recall=metrics.recall_score(y_true=labels, y_pred=rounded_preds),
            f1_score=metrics.f1_score(y_true=labels, y_pred=rounded_preds),
            roc_auc=metrics.roc_auc_score(y_true=labels, y_score=predictions),
        )

        return metrics



import tensorflow as tf



# TODO: This is a band-aid around unimplemented support for sparse gradients in TF2.0
# We only use it in a situation where the gradients are always covering all inputs
# (and hence sparse/dense has little difference), so this is not a substantial
# performance hit, just inconvenient code.
@tf.custom_gradient
def gather_dense_gradient(params, indices):
    def grad(ys):
        return tf.scatter_nd(tf.expand_dims(indices, 1), ys, tf.shape(params)), None

    return tf.gather(params, indices), grad


"""General dataset class for datasets stored as JSONLines files."""
import logging
from typing import Any, Dict, Iterator, List, Optional, Tuple, Set

import numpy as np
from dpu_utils.utils import RichPath

from .graph_dataset import DataFold, GraphDataset, GraphSampleType, GraphSample
from .utils import compute_number_of_edge_types, get_tied_edge_types, process_adjacency_lists

logger = logging.getLogger(__name__)


class JsonLGraphDataset(GraphDataset[GraphSampleType]):
    """
    General class representing pre-split datasets in JSONLines format.
    Concretely, this class expects the following:
    * In the data directory, files "train.jsonl.gz", "valid.jsonl.gz" and
      "test.jsonl.gz" are used to store the train/valid/test datasets.
    * Each of the files is gzipped text file in which each line is a valid
      JSON dictionary with a "graph" key, which in turn points to a
      dictionary with keys
       - "node_features" (list of numerical initial node labels)
       - "adjacency_lists" (list of list of directed edge pairs)
    """

    @classmethod
    def get_default_hyperparameters(cls) -> Dict[str, Any]:
        super_hypers = super().get_default_hyperparameters()
        this_hypers = {
            "num_fwd_edge_types": 3,
            "add_self_loop_edges": True,
            "tie_fwd_bkwd_edges": True,
        }
        super_hypers.update(this_hypers)
        return super_hypers

    def __init__(
        self, params: Dict[str, Any], metadata: Optional[Dict[str, Any]] = None, **kwargs,
    ):
        super().__init__(params, metadata=metadata, **kwargs)
        self._params = params
        self._num_fwd_edge_types = params["num_fwd_edge_types"]

        self._tied_fwd_bkwd_edge_types = get_tied_edge_types(
            tie_fwd_bkwd_edges=params["tie_fwd_bkwd_edges"],
            num_fwd_edge_types=params["num_fwd_edge_types"],
        )

        self._num_edge_types = compute_number_of_edge_types(
            tied_fwd_bkwd_edge_types=self._tied_fwd_bkwd_edge_types,
            num_fwd_edge_types=self._num_fwd_edge_types,
            add_self_loop_edges=params["add_self_loop_edges"],
        )

        self._loaded_data: Dict[DataFold, List[GraphSampleType]] = {}

    @property
    def num_edge_types(self) -> int:
        return self._num_edge_types

    @property
    def node_feature_shape(self) -> Tuple:
        """Return the shape of the node features."""
        node_feature_shape = self.metadata.get("_node_feature_shape")
        if node_feature_shape is None:
            some_data_fold = next(iter(self._loaded_data.values()))
            node_feature_shape = (len(some_data_fold[0].node_features[0]),)
            self.metadata["_node_feature_shape"] = node_feature_shape
        return node_feature_shape

    def load_metadata(self, path: RichPath) -> None:
        """Load the metadata for a dataset (such as vocabularies, names of properties, ...)
        from a path on disk.

        Note: Implementors needing to act on metadata information before loading any actual data
        should override this method.
        """
        if self.metadata == {}:
            metadata_path = path.join("metadata.pkl.gz")
            if metadata_path.exists():
                logger.info(f"Loading metadata from {metadata_path}")
                self._metadata = metadata_path.read_by_file_suffix()
        else:
            logger.warning("Using metadata passed to constructor, not metadata stored with data.")

    def load_data(self, path: RichPath, folds_to_load: Optional[Set[DataFold]] = None) -> None:
        """Load the data from disk."""
        logger.info(f"Starting to load data from {path}.")
        self.load_metadata(path)

        # If we haven't defined what folds to load, load all:
        if folds_to_load is None:
            folds_to_load = {DataFold.TRAIN, DataFold.VALIDATION, DataFold.TEST}

        if DataFold.TRAIN in folds_to_load:
            self._loaded_data[DataFold.TRAIN] = self.__load_data(path.join("train.jsonl.gz"))
            logger.debug("Done loading training data.")
        if DataFold.VALIDATION in folds_to_load:
            self._loaded_data[DataFold.VALIDATION] = self.__load_data(path.join("valid.jsonl.gz"))
            logger.debug("Done loading validation data.")
        if DataFold.TEST in folds_to_load:
            self._loaded_data[DataFold.TEST] = self.__load_data(path.join("test.jsonl.gz"))
            logger.debug("Done loading test data.")

    def load_data_from_list(
        self, datapoints: List[Dict[str, Any]], target_fold: DataFold = DataFold.TEST
    ):
        if target_fold not in self._loaded_data:
            self._loaded_data[target_fold] = []
        for datapoint in datapoints:
            self._loaded_data[target_fold].append(self._process_raw_datapoint(datapoint))

    def __load_data(self, data_file: RichPath) -> List[GraphSampleType]:
        return [
            self._process_raw_datapoint(datapoint) for datapoint in data_file.read_by_file_suffix()
        ]

    def _process_raw_datapoint(self, datapoint: Dict[str, Any]) -> GraphSampleType:
        node_features = datapoint["graph"]["node_features"]
        type_to_adj_list, type_to_num_incoming_edges = self._process_raw_adjacency_lists(
            raw_adjacency_lists=datapoint["graph"]["adjacency_lists"],
            num_nodes=len(node_features),
        )

        return GraphSample(
            adjacency_lists=type_to_adj_list,
            type_to_node_to_num_inedges=type_to_num_incoming_edges,
            node_features=node_features,
        )

    def _process_raw_adjacency_lists(
        self, raw_adjacency_lists: List[List[Tuple]], num_nodes: int
    ) -> Tuple[List[np.ndarray], np.ndarray]:
        return process_adjacency_lists(
            adjacency_lists=raw_adjacency_lists,
            num_nodes=num_nodes,
            add_self_loop_edges=self.params["add_self_loop_edges"],
            tied_fwd_bkwd_edge_types=self._tied_fwd_bkwd_edge_types,
        )

    def _graph_iterator(self, data_fold: DataFold) -> Iterator[GraphSampleType]:
        if data_fold == DataFold.TRAIN:
            np.random.shuffle(self._loaded_data[data_fold])
        return iter(self._loaded_data[data_fold])



from .message_passing import MessagePassing, MessagePassingInput, MESSAGE_PASSING_IMPLEMENTATIONS
from .rgat import RGAT
from .rgcn import RGCN
from .rgin import RGIN
from .ggnn import GGNN
from .gnn_edge_mlp import GNN_Edge_MLP
from .gnn_film import GNN_FiLM

def get_message_passing_class(message_calculation_class_name: str):
    calculation_class = MESSAGE_PASSING_IMPLEMENTATIONS.get(message_calculation_class_name.lower())
    if calculation_class is None:
        raise ValueError(f"Unknown message passing type: {message_calculation_class_name}")
    return calculation_class

def get_known_message_passing_classes():
    for message_passing_implementation in MESSAGE_PASSING_IMPLEMENTATIONS.values():
        yield message_passing_implementation.__name__



import time
from abc import abstractmethod
from typing import Tuple, List, Dict, Optional, Any, Iterable, Union

import tensorflow as tf

from tf2_gnn import GNNInput, GNN
from tf2_gnn.data import GraphDataset
from tf2_gnn.utils.polynomial_warmup_and_decay_schedule import (
    PolynomialWarmupAndDecaySchedule,
)


class GraphTaskModel(tf.keras.Model):
    @classmethod
    def get_default_hyperparameters(cls, mp_style: Optional[str] = None) -> Dict[str, Any]:
        """Get the default hyperparameter dictionary for the class."""
        params = {f"gnn_{name}": value for name, value in GNN.get_default_hyperparameters(mp_style).items()}
        these_hypers: Dict[str, Any] = {
            "optimizer": "Adam",  # One of "SGD", "RMSProp", "Adam"
            "learning_rate": 0.001,
            "learning_rate_warmup_steps": None,
            "learning_rate_decay_steps": None,
            "momentum": 0.85,
            "rmsprop_rho": 0.98,  # decay of gradients in RMSProp (unused otherwise)
            "gradient_clip_value": None,  # Set to float value to clip each gradient separately
            "gradient_clip_norm": None,  # Set to value to clip gradients by their norm
            "gradient_clip_global_norm": None,  # Set to value to clip gradients by their global norm
            "use_intermediate_gnn_results": False,
        }
        params.update(these_hypers)
        return params

    def __init__(
        self,
        params: Dict[str, Any],
        dataset: GraphDataset,
        name: str = None,
        disable_tf_function_build: bool = False
    ):
        super().__init__(name=name)
        self._params = params
        self._num_edge_types = dataset.num_edge_types
        self._use_intermediate_gnn_results = params.get(
            "use_intermediate_gnn_results", False
        )
        self._disable_tf_function_build = disable_tf_function_build

        # Keep track of the training step as a TF variable
        self._train_step_counter = tf.Variable(
            initial_value=0, trainable=False, dtype=tf.int32, name="training_step"
        )

        # Store a couple of descriptions for jit compilation in the build function.
        batch_description = dataset.get_batch_tf_data_description()
        self._batch_feature_names = tuple(batch_description.batch_features_types.keys())
        self._batch_label_names = tuple(batch_description.batch_labels_types.keys())
        self._batch_feature_spec = tuple(
            tf.TensorSpec(
                shape=batch_description.batch_features_shapes[name],
                dtype=batch_description.batch_features_types[name],
            )
            for name in self._batch_feature_names
        )
        self._batch_label_spec = tuple(
            tf.TensorSpec(
                shape=batch_description.batch_labels_shapes[name],
                dtype=batch_description.batch_labels_types[name],
            )
            for name in self._batch_label_names
        )

    @staticmethod
    def _pack(input: Dict[str, Any], names: Tuple[str, ...]) -> Tuple:
        return tuple(input.get(name) for name in names)

    def _pack_features(self, batch_features: Dict[str, Any]) -> Tuple:
        return self._pack(batch_features, self._batch_feature_names)

    def _pack_labels(self, batch_labels: Dict[str, Any]) -> Tuple:
        return self._pack(batch_labels, self._batch_label_names)

    @staticmethod
    def _unpack(input: Tuple, names: Tuple) -> Dict[str, Any]:
        return {name: value for name, value in zip(names, input) if value is not None}

    def _unpack_features(self, batch_features: Tuple) -> Dict[str, Any]:
        return self._unpack(batch_features, self._batch_feature_names)

    def _unpack_labels(self, batch_labels: Tuple) -> Dict[str, Any]:
        return self._unpack(batch_labels, self._batch_label_names)

    def build(self, input_shapes: Dict[str, Any]):
        graph_params = {
            name[4:]: value for name, value in self._params.items() if name.startswith("gnn_")
        }
        self._gnn = GNN(graph_params)
        self._gnn.build(
            GNNInput(
                node_features=self.get_initial_node_feature_shape(input_shapes),
                adjacency_lists=tuple(
                    input_shapes[f"adjacency_list_{edge_type_idx}"]
                    for edge_type_idx in range(self._num_edge_types)
                ),
                node_to_graph_map=tf.TensorShape((None,)),
                num_graphs=tf.TensorShape(()),
            )
        )

        super().build([])

        if not self._disable_tf_function_build:
            setattr(
                self,
                "_fast_run_step",
                tf.function(
                    input_signature=(
                        self._batch_feature_spec,
                        self._batch_label_spec,
                        tf.TensorSpec(shape=(), dtype=tf.bool),
                    )
                )(self._fast_run_step),
            )

    def get_initial_node_feature_shape(self, input_shapes) -> tf.TensorShape:
        return input_shapes["node_features"]

    def compute_initial_node_features(self, inputs, training: bool) -> tf.Tensor:
        return inputs["node_features"]

    @abstractmethod
    def compute_task_output(
        self,
        batch_features: Dict[str, tf.Tensor],
        final_node_representations: Union[tf.Tensor, Tuple[tf.Tensor, Tuple[tf.Tensor, ...]]],
        training: bool,
    ) -> Any:
        """Compute task-specific output (labels, scores, regression values, ...).

        Args:
            batch_features: Input data for minibatch (as generated by the used datasets
                _finalise_batch method).
            final_node_representations:
                Per default (or if the hyperparameter "use_intermediate_gnn_results" was
                set to False), the final representations of the graph nodes as computed
                by the GNN.
                If the hyperparameter "use_intermediate_gnn_results" was set to True,
                a pair of the final node representation and all intermediate node
                representations, including the initial one.
            training: Flag indicating if we are training or not.

        Returns:
            Implementor's choice, but will be passed as task_output to compute_task_metrics
            during training/evaluation.
        """
        pass

    def compute_final_node_representations(self, inputs, training: bool):
        # Pack input data from keys back into a tuple:
        adjacency_lists: Tuple[tf.Tensor, ...] = tuple(
            inputs[f"adjacency_list_{edge_type_idx}"]
            for edge_type_idx in range(self._num_edge_types)
        )

        # Start the model computations:
        initial_node_features = self.compute_initial_node_features(inputs, training)
        gnn_input = GNNInput(
            node_features=initial_node_features,
            adjacency_lists=adjacency_lists,
            node_to_graph_map=inputs["node_to_graph_map"],
            num_graphs=inputs["num_graphs_in_batch"],
        )

        gnn_output = self._gnn(
            gnn_input,
            training=training,
            return_all_representations=self._use_intermediate_gnn_results
        )
        return gnn_output

    def call(self, inputs, training: bool):
        final_node_representations = self.compute_final_node_representations(inputs, training)
        return self.compute_task_output(inputs, final_node_representations, training)

    @abstractmethod
    def compute_task_metrics(
        self,
        batch_features: Dict[str, tf.Tensor],
        task_output: Any,
        batch_labels: Dict[str, tf.Tensor],
    ) -> Dict[str, tf.Tensor]:
        """Compute task-specific loss & metrics (accuracy, F1 score, ...)

        Args:
            batch_features: Input data for minibatch (as generated by the used datasets
                _finalise_batch method).
            task_output: Output generated by compute_task_output.
            batch_labels: Target labels for minibatch (as generated by the used datasets
                _finalise_batch method).

        Returns:
            Dictionary of different metrics. Has to contain value for key
            "loss" (which will be used during training as starting point for backprop).
        """
        pass

    @abstractmethod
    def compute_epoch_metrics(self, task_results: List[Any]) -> Tuple[float, str]:
        """Compute single value used to measure quality of model at one epoch, where
        lower is better.
        This value, computed on the validation set, is used to determine if model
        training is still improving results.

        Args:
            task_results: List of results obtained by compute_task_metrics for the
                batches in one epoch.

        Returns:
            Pair of a metric value (lower ~ better) and a human-readable string
            describing it.
        """
        pass

    def _make_optimizer(
        self,
        learning_rate: Optional[
            Union[float, tf.keras.optimizers.schedules.LearningRateSchedule]
        ] = None,
    ) -> tf.keras.optimizers.Optimizer:
        """
        Create fresh optimizer.

        Args:
            learning_rate: Optional setting for learning rate; if unset, will
                use value from self._params["learning_rate"].
        """
        if learning_rate is None:
            learning_rate = self._params["learning_rate"]

            num_warmup_steps = self._params.get("learning_rate_warmup_steps")
            num_decay_steps = self._params.get("learning_rate_decay_steps")
            if num_warmup_steps is not None or num_decay_steps is not None:
                initial_learning_rate = 0.00001
                final_learning_rate = 0.00001
                if num_warmup_steps is None:
                    num_warmup_steps = -1  # Make sure that we have no warmup phase
                    initial_learning_rate = learning_rate
                if num_decay_steps is None:
                    num_decay_steps = (
                        1  # Value doesn't matter, but needs to be non-zero
                    )
                    final_learning_rate = learning_rate
                learning_rate = PolynomialWarmupAndDecaySchedule(
                    learning_rate=learning_rate,
                    warmup_steps=num_warmup_steps,
                    decay_steps=num_decay_steps,
                    initial_learning_rate=initial_learning_rate,
                    final_learning_rate=final_learning_rate,
                    power=1.0,
                )

        optimizer_name = self._params["optimizer"].lower()
        if optimizer_name == "sgd":
            return tf.keras.optimizers.SGD(
                learning_rate=learning_rate, momentum=self._params["momentum"],
            )
        elif optimizer_name == "rmsprop":
            return tf.keras.optimizers.RMSprop(
                learning_rate=learning_rate,
                momentum=self._params["momentum"],
                rho=self._params["rmsprop_rho"],
            )
        elif optimizer_name == "adam":
            return tf.keras.optimizers.Adam(learning_rate=learning_rate,)
        else:
            raise Exception('Unknown optimizer "%s".' % (self._params["optimizer"]))

    def _apply_gradients(
        self, gradient_variable_pairs: Iterable[Tuple[tf.Tensor, tf.Variable]]
    ) -> None:
        """
        Apply gradients to the models variables during training.

        Args:
            gradient_variable_pairs: Iterable of pairs of gradients for a variable
                and the variable itself. Suitable to be fed into
                tf.keras.optimizer.*.apply_gradients.
        """
        if getattr(self, "_optimizer", None) is None:
            self._optimizer = self._make_optimizer()

        # Filter out variables without gradients:
        gradient_variable_pairs = [
            (grad, var) for (grad, var) in gradient_variable_pairs if grad is not None
        ]
        clip_val = self._params.get("gradient_clip_value")
        clip_norm_val = self._params.get("gradient_clip_norm")
        clip_global_norm_val = self._params.get("gradient_clip_global_norm")

        if clip_val is not None:
            if clip_norm_val is not None:
                raise ValueError("Both 'gradient_clip_value' and 'gradient_clip_norm' are set, but can only use one at a time.")
            if clip_global_norm_val is not None:
                raise ValueError("Both 'gradient_clip_value' and 'gradient_clip_global_norm' are set, but can only use one at a time.")
            gradient_variable_pairs = [
                (tf.clip_by_value(grad, -clip_val, clip_val), var)
                for (grad, var) in gradient_variable_pairs
            ]
        elif clip_norm_val is not None:
            if clip_global_norm_val is not None:
                raise ValueError("Both 'gradient_clip_norm' and 'gradient_clip_global_norm' are set, but can only use one at a time.")
            gradient_variable_pairs = [
                (tf.clip_by_norm(grad, clip_norm_val), var)
                for (grad, var) in gradient_variable_pairs
            ]
        elif clip_global_norm_val is not None:
            grads = [grad for (grad, _) in gradient_variable_pairs]
            clipped_grads, _ = tf.clip_by_global_norm(grads, clip_norm=clip_global_norm_val)
            gradient_variable_pairs = [
                (clipped_grad, var)
                for (clipped_grad, (_, var)) in zip(clipped_grads, gradient_variable_pairs)
            ]

        self._optimizer.apply_gradients(gradient_variable_pairs)

    # ----------------------------- Training Loop
    def _run_step(
        self,
        batch_features: Dict[str, tf.Tensor],
        batch_labels: Dict[str, tf.Tensor],
        training: bool,
    ) -> Dict[str, tf.Tensor]:
        batch_features_tuple = self._pack_features(batch_features)
        batch_labels_tuple = self._pack_labels(batch_labels)

        return self._fast_run_step(batch_features_tuple, batch_labels_tuple, training)

    def _fast_run_step(
        self,
        batch_features_tuple: Tuple[tf.Tensor],
        batch_labels_tuple: Tuple[tf.Tensor],
        training: bool,
    ):
        batch_features = self._unpack_features(batch_features_tuple)
        batch_labels = self._unpack_labels(batch_labels_tuple)

        with tf.GradientTape() as tape:
            task_output = self(batch_features, training=training)
            task_metrics = self.compute_task_metrics(
                batch_features=batch_features,
                task_output=task_output,
                batch_labels=batch_labels,
            )

        def _training_update():
            gradients = tape.gradient(task_metrics["loss"], self.trainable_variables)
            self._apply_gradients(zip(gradients, self.trainable_variables))
            self._train_step_counter.assign_add(1)

        def _no_op():
            pass

        tf.cond(training, true_fn=_training_update, false_fn=_no_op)

        return task_metrics

    def run_one_epoch(
        self, dataset: tf.data.Dataset, quiet: bool = False, training: bool = True,
    ) -> Tuple[float, float, List[Any]]:
        epoch_time_start = time.time()
        total_num_graphs = 0
        task_results = []
        total_loss = tf.constant(0, dtype=tf.float32)
        for step, (batch_features, batch_labels) in enumerate(dataset):
            task_metrics = self._run_step(batch_features, batch_labels, training)
            total_loss += task_metrics["loss"]
            total_num_graphs += batch_features["num_graphs_in_batch"]
            task_results.append(task_metrics)

            if not quiet:
                epoch_graph_average_loss = (
                    total_loss / float(total_num_graphs)
                ).numpy()
                batch_graph_average_loss = task_metrics["loss"] / float(
                    batch_features["num_graphs_in_batch"]
                )
                steps_per_second = step / (time.time() - epoch_time_start)
                print(
                    f"   Step: {step:4d}"
                    f"  |  Epoch graph avg. loss = {epoch_graph_average_loss:.5f}"
                    f"  |  Batch graph avg. loss = {batch_graph_average_loss:.5f}"
                    f"  |  Steps per sec = {steps_per_second:.5f}",
                    end="\r"
                )
        if not quiet:
            print("\r\x1b[K", end="")
        total_time = time.time() - epoch_time_start
        return total_loss / float(total_num_graphs), float(total_num_graphs) / total_time, task_results

    # ----------------------------- Prediction Loop
    def predict(self, dataset: tf.data.Dataset):
        task_outputs = []
        for batch_features, _ in dataset:
            task_outputs.append(self(batch_features, training=False))

        # Note: This assumes that the task output is a tensor (true for classification, regression,
        #  etc.) but subclasses implementing more complicated outputs will need to override this.
        return tf.concat(task_outputs, axis=0)

    def evaluate_model(self, dataset: tf.data.Dataset) -> Dict[str, float]:
        """Evaluate the model using metrics that make sense for its application area.

        Args:
            dataset: A dataset to evaluate on, same format as used in the training loop.

        Returns:
            Dictionary mapping metric names (e.g., "accuracy", "roc_auc") to their respective
            values.
        """
        raise NotImplementedError()


#!/usr/bin/env python
"""Make the module pip installable."""

import setuptools
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="tf2_gnn",
    version="2.8.0",
    license="MIT",
    author="Marc Brockschmidt",
    author_email="mabrocks@microsoft.com",
    description="TensorFlow 2.0 implementation of Graph Neural Networks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/tf2-gnn/",
    python_requires=">=3.6",
    install_requires=[
        "numpy",
        "tensorflow>=2.0.0",
        "dpu-utils>=0.2.7",
        "h5py",
    ],
    packages=setuptools.find_packages(where="."),
    package_dir={"": "."},
    package_data={"": ["default_hypers/*.json"]},
    entry_points={
        "console_scripts": [
            "tf2_gnn_train = tf2_gnn.cli.train:run",
            "tf2_gnn_test = tf2_gnn.cli.test:run",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from copy import deepcopy
from typing import List, Set, Tuple, Union

import numpy as np

Edge = Tuple[int, int]


def process_adjacency_lists(
    adjacency_lists: List[List[Edge]],
    num_nodes: int,
    add_self_loop_edges: bool,
    tied_fwd_bkwd_edge_types: Set[int],
    self_loop_edge_type: int = 0,
) -> Tuple[List[np.ndarray], np.ndarray]:
    """Process adjacency lists by adding backward edges and self loops.

    Args:
        adjacency_lists: adjacency lists as a list of lists, with one list per edge type.
        num_nodes: number of nodes in the graph.
        add_self_loop_edges: whether to add self loops.
        tied_fwd_bkwd_edge_types: For these forward edge types, the added backward edges will have
            the same type as the forward edge. For all remaining forward edge types, the backward
            edges will get a new fresh edge type.
        self_loop_edge_type: edge type to use for the self loops. Also supports negative numbers:
            for example, setting this to -1 will result in self loop edge type being the last one.
            Only has effect if `add_self_loop_edges` is set.

    Returns:
        Processed adjacency lists (with backward edges and self loops added, and each inner list
        converted to numpy array), and an array of shape [num_total_edge_types, num_nodes]
        containing counts of edges of a given type adjacent to a given node.
    """
    adjacency_lists = _add_backward_edges(adjacency_lists, tied_fwd_bkwd_edge_types)

    # Add self loops after adding backward edges to avoid adding loops twice.
    if add_self_loop_edges:
        num_edge_types = len(adjacency_lists)

        edge_type_lb = -(num_edge_types + 1)
        edge_type_ub = num_edge_types

        assert (
            edge_type_lb <= self_loop_edge_type <= edge_type_ub
        ), "Self loop edge type {} should be in range [{}, {}].".format(
            self_loop_edge_type, edge_type_lb, edge_type_ub
        )

        if self_loop_edge_type < 0:
            self_loop_edge_type += num_edge_types + 1

        adjacency_lists = _add_self_loop_edges(adjacency_lists, num_nodes, self_loop_edge_type)

    type_to_num_incoming_edges = _compute_type_to_num_inedges(
        adjacency_lists=adjacency_lists, num_nodes=num_nodes
    )

    return _convert_adjacency_lists_to_numpy_arrays(adjacency_lists), type_to_num_incoming_edges


def get_tied_edge_types(
    tie_fwd_bkwd_edges: Union[bool, List[int]], num_fwd_edge_types: int
) -> Set[int]:
    """Get the forward edge types which should be tied with their respective backward edge types.

    Args:
        tie_fwd_bkwd_edges: either an explicit list of edge types to tie (in which case that list is
            returned as a set), or a bool value (whether to tie all edge types, or none).

    Returns:
        Set of forward edge types to tie, which can be passed to`process_adjacency_lists`.
    """
    if isinstance(tie_fwd_bkwd_edges, list):
        return set(tie_fwd_bkwd_edges)
    elif tie_fwd_bkwd_edges:
        return set(range(num_fwd_edge_types))
    else:
        return {}


def compute_number_of_edge_types(
    tied_fwd_bkwd_edge_types: Set[int], num_fwd_edge_types: int, add_self_loop_edges: bool
) -> int:
    """Computes the number of edge types after adding backward edges and possibly self loops."""
    return 2 * num_fwd_edge_types - len(tied_fwd_bkwd_edge_types) + int(add_self_loop_edges)


def _add_self_loop_edges(
    adjacency_lists: List[List[Edge]], num_nodes: int, self_loop_edge_type: int = 0
) -> List[List[Edge]]:
    self_loops = [(i, i) for i in range(num_nodes)]

    adjacency_lists = deepcopy(adjacency_lists)
    adjacency_lists.insert(self_loop_edge_type, self_loops)

    return adjacency_lists


def _add_backward_edges(
    adjacency_lists: List[List[Edge]], tied_fwd_bkwd_edge_types: Set[int]
) -> List[List[Edge]]:
    # Make sure the output will contain newly created lists.
    new_adjacency_lists = deepcopy(adjacency_lists)

    for edge_type in range(len(adjacency_lists)):
        flipped_adjacency_list = [(dest, src) for (src, dest) in adjacency_lists[edge_type]]

        if edge_type in tied_fwd_bkwd_edge_types:
            new_adjacency_lists[edge_type] += flipped_adjacency_list
        else:
            new_adjacency_lists.append(flipped_adjacency_list)

    return new_adjacency_lists


def _compute_type_to_num_inedges(adjacency_lists: List[List[Edge]], num_nodes: int) -> np.ndarray:
    num_edge_types = len(adjacency_lists)
    type_to_num_incoming_edges = np.zeros(shape=(num_edge_types, num_nodes))

    for edge_type, edges in enumerate(adjacency_lists):
        for _, dest in edges:
            type_to_num_incoming_edges[edge_type, dest] += 1

    return type_to_num_incoming_edges


def _convert_adjacency_lists_to_numpy_arrays(adjacency_lists: List[List[Edge]]) -> List[np.ndarray]:
    return [
        np.array(adj_list, dtype=np.int32)
        if len(adj_list) > 0
        else np.zeros(shape=(0, 2), dtype=np.int32)
        for adj_list in adjacency_lists
    ]
